"""

Client class is a wrapper around the REST API of the exchange. It provides methods to interact with the exchange API.
"""

from decimal import Decimal
import time
from typing import Any, List

import eth_account
import requests
from eip712_structs import make_domain
from eth_account.messages import encode_structured_data
from web3 import Web3
from web3.exceptions import TransactionNotFound

from hundred_x.constants import APIS, CONTRACTS, LOGIN_MESSAGE
from hundred_x.eip_712 import CancelOrder, CancelOrders, LoginMessage, Order, Withdraw
from hundred_x.enums import ApiType, Environment, OrderSide, OrderType, TimeInForce
from hundred_x.utils import from_message_to_payload, get_abi

headers = {
    "Accept": "application/json",
    "Content-Type": "application/json",
}


PROTOCOL_ABI = get_abi("protocol")
ERC_20_ABI = get_abi("erc20")


class HundredXClient:
    def __init__(
        self,
        env: Environment = Environment.TESTNET,
        private_key: str = None,
        subaccount_id: int = 0,
    ):
        """
        Initialize the client with the given environment.
        """
        self.env = env
        self.rest_url = APIS[env][ApiType.REST]
        self.websocket_url = APIS[env][ApiType.WEBSOCKET]
        self.wallet = eth_account.Account.from_key(private_key)
        self.public_key = self.wallet.address
        self.subaccount_id = subaccount_id
        self.session_cookie = {}
        self.web3 = Web3(Web3.HTTPProvider("https://sepolia.blast.io"))
        self.domain = make_domain(
            name="100x",
            version="0.0.0",
            chainId=CONTRACTS[env]["CHAIN_ID"],
            verifyingContract=CONTRACTS[env]["VERIFYING_CONTRACT"],
        )

    def _current_timestamp(self):
        timestamp_ms = int(time.time() * 1000)
        return timestamp_ms

    def generate_and_sign_message(self, message_class, **kwargs):
        """
        Generate and sign a message.
        """
        message = message_class(**kwargs)
        message = message.to_message(self.domain)
        signable_message = encode_structured_data(message)
        signed = self.wallet.sign_message(signable_message)
        message["message"]["signature"] = signed.signature.hex()
        return message["message"]

    def get_shared_params(self, asset: str = None, subaccount_id: int = None):
        params = {
            "account": self.public_key,
        }
        if asset is not None:
            params["asset"] = self.get_contract_address(asset)
        if subaccount_id is not None:
            params["subAccountId"] = subaccount_id
        return params

    def send_message_to_endpoint(self, endpoint: str, method: str, message: dict, authenticated: bool = True):
        """
        Send a message to an endpoint.
        """
        payload = from_message_to_payload(message)
        response = requests.request(
            method,
            self.rest_url + endpoint,
            headers={} if not authenticated else self.authenticated_headers,
            json=payload,
        )
        if response.status_code != 200:
            raise Exception(f"Failed to send message: {response.text} {response.status_code} {self.rest_url} {payload}")
        return response.json()

    def withdraw(self, subaccount_id: int, quantity: int, asset: str = "USDB"):
        """
        Generate a withdrawal message and sign it.
        """

        message = self.generate_and_sign_message(
            Withdraw,
            quantity=int(quantity * 1e18),
            nonce=self._current_timestamp(),
            **self.get_shared_params(subaccount_id=subaccount_id, asset=asset),
        )
        return self.send_message_to_endpoint("/v1/withdraw", "POST", message)

    def create_order(
        self,
        subaccount_id: int,
        product_id: int,
        quantity: int,
        price: int,
        side: OrderSide,
        order_type: OrderType,
        time_in_force: TimeInForce,
        nonce: int = 0,
    ):
        """
        Create an order.
        """
        ts = self._current_timestamp()
        if nonce == 0:
            nonce = ts
        message = self.generate_and_sign_message(
            Order,
            subAccountId=subaccount_id,
            productId=product_id,
            quantity=int(Decimal(str(quantity)) * Decimal(1e18)),
            price=int(Decimal(str(price)) * Decimal(1e18)),
            isBuy=side.value,
            orderType=order_type.value,
            timeInForce=time_in_force.value,
            nonce=nonce,
            expiration=(ts + 1000 * 60 * 60 * 24) * 1000,
            **self.get_shared_params(),
        )
        return self.send_message_to_endpoint("/v1/order", "POST", message)

    def cancel_and_replace_order(
        self,
        subaccount_id: int,
        product_id: int,
        quantity: int,
        price: int,
        side: OrderSide,
        order_type: OrderType,
        time_in_force: TimeInForce,
        order_id_to_cancel: str,
    ):
        """
        Cancel and replace an order.
        """
        ts = self._current_timestamp()
        message = self.generate_and_sign_message(
            Order,
            subAccountId=subaccount_id,
            productId=product_id,
            quantity=int(Decimal(str(quantity)) * Decimal(1e18)),
            price=int(Decimal(str(price)) * Decimal(1e18)),
            isBuy=side.value,
            orderType=order_type.value,
            timeInForce=time_in_force.value,
            nonce=ts,
            expiration=(ts + 1000 * 60 * 60 * 24) * 1000,
            **self.get_shared_params(),
        )
        message["orderIdToCancel"] = order_id_to_cancel
        return self.send_message_to_endpoint("/v1/cancel-and-replace", "PUT", message)

    def cancel_order(self, subaccount_id: int, product_id: int, order_id: int):
        """
        Cancel an order.
        """
        message = self.generate_and_sign_message(
            CancelOrder,
            subAccountId=subaccount_id,
            productId=product_id,
            orderId=order_id,
            **self.get_shared_params(),
        )
        return self.send_message_to_endpoint("/v1/order", "DELETE", message)

    def cancel_all_orders(self, subaccount_id: int, product_id: int):
        """
        Cancel all orders.
        """
        message = self.generate_and_sign_message(
            CancelOrders,
            subAccountId=subaccount_id,
            productId=product_id,
            **self.get_shared_params(),
        )
        return self.send_message_to_endpoint("/v1/openOrders", "DELETE", message)

    def create_authenticated_session_with_service(self):
        login_payload = self.generate_and_sign_message(
            LoginMessage,
            message=LOGIN_MESSAGE,
            timestamp=self._current_timestamp(),
            **self.get_shared_params(),
        )
        response = self.send_message_to_endpoint("/v1/session/login", "POST", login_payload, authenticated=False)
        self.session_cookie = response.get("value")
        return response

    def list_products(self) -> List[Any]:
        """
        Get a list of all available products.
        """
        return requests.get(self.rest_url + "/v1/products").json()

    def get_product(self, product_symbol: str) -> Any:
        """
        Get the details of a specific product.
        """
        return requests.get(self.rest_url + f"/v1/products/{product_symbol}").json()

    def get_trade_history(self, symbol: str, lookback: int) -> Any:
        """
        Get the trade history for a specific product symbol and lookback amount.
        """
        return requests.get(
            self.rest_url + "/v1/trade-history",
            params={"symbol": symbol, "lookback": lookback},
        ).json()

    def get_server_time(self) -> Any:
        """
        Get the server time.
        """
        return requests.get(self.rest_url + "/v1/time").json()

    def get_candlestick(self, symbol: str, **kwargs) -> Any:
        """
        Get the candlestick data for a specific product.
        """
        params = {"symbol": symbol}
        for arg in ["interval", "start_time", "end_time", "limit"]:
            var = kwargs.get(arg)
            if var is not None:
                params[arg] = var
        return requests.get(
            self.rest_url + "/v1/uiKlines",
            params=params,
        ).json()

    def get_symbol(self, symbol: str) -> Any:
        """
        Get the details of a specific symbol.
        """
        endpoint = f"/v1/ticker/24hr?symbol={symbol}"
        return requests.get(self.rest_url + endpoint).json()[0]

    def get_depth(self, symbol: str, **kwargs) -> Any:
        """
        Get the depth data for a specific product.
        """
        params = {"symbol": symbol}
        for arg in ["limit"]:
            var = kwargs.get(arg)
            if var is not None:
                params[arg] = var
        return requests.get(
            self.rest_url + "/v1/depth",
            params=params,
        ).json()

    def login(self):
        """
        Login to the exchange.
        """
        response = self.create_authenticated_session_with_service()
        if response is None:
            raise Exception("Failed to login")

    def get_session_status(self):
        """
        Get the current session status.
        """
        return requests.get(self.rest_url + "/v1/session/status", headers=self.authenticated_headers).json()

    @property
    def authenticated_headers(self):
        return {
            "cookie": f"connectedAddress={self.session_cookie}",
        }

    def logout(self):
        """
        Logout from the exchange.
        """
        return requests.get(self.rest_url + "/v1/session/logout", headers=self.authenticated_headers).json()

    def get_spot_balances(self):
        """
        Get the spot balances.
        """
        return requests.get(
            self.rest_url + "/v1/balances",
            headers=self.authenticated_headers,
            params={"account": self.public_key, "subAccountId": self.subaccount_id},
        ).json()

    def get_position(self, symbol: str):
        """
        Get the position for a specific symbol.
        """
        return requests.get(
            self.rest_url + "/v1/positionRisk",
            headers=self.authenticated_headers,
            params={
                "symbol": symbol,
                "account": self.public_key,
                "subAccountId": self.subaccount_id,
            },
        ).json()

    def get_approved_signers(self):
        """
        Get the approved signers.
        """
        return requests.get(
            self.rest_url + "/v1/approved-signers",
            headers=self.authenticated_headers,
            params={"account": self.public_key, "subAccountId": self.subaccount_id},
        ).json()

    def get_open_orders(
        self,
        symbol: str = None,
    ):
        """
        Get the open orders.
        """
        params = {"account": self.public_key, "subAccountId": self.subaccount_id}
        if symbol is not None:
            params["symbol"] = symbol
        return requests.get(
            self.rest_url + "/v1/openOrders",
            headers=self.authenticated_headers,
            params=params,
        ).json()

    def get_orders(self, symbol: str = None, ids: List[str] = None):
        """
        Get the open orders.
        """
        params = {"account": self.public_key, "subAccountId": self.subaccount_id}

        if ids is not None:
            params["ids"] = ids
        if symbol is not None:
            params["symbol"] = symbol

        response = requests.get(
            self.rest_url + "/v1/orders",
            headers=self.authenticated_headers,
            params=params,
        )
        if response.status_code != 200:
            raise Exception(
                f"Failed to get orders: {response.text} {response.status_code} " + f"{self.rest_url} {params}"
            )
        return response.json()

    def deposit(self, subaccount_id: int, quantity: int, asset: str = "USDB"):
        """
        Deposit an asset.
        """
        # we need to check if we have sufficient balance to deposit
        required_wei = int(Decimal(str(quantity)) * Decimal(1e18))
        # we check the approvals
        asset_contract = self.get_contract(asset)

        approved_amount = asset_contract.functions.allowance(
            self.public_key, self.get_contract_address("PROTOCOL")
        ).call()
        if approved_amount < required_wei:
            txn = asset_contract.functions.approve(
                self.get_contract_address("PROTOCOL"), required_wei
            ).build_transaction(
                {
                    "from": self.public_key,
                    "nonce": self.web3.eth.get_transaction_count(self.public_key),
                }
            )
            signed_txn = self.wallet.sign_transaction(txn)
            result = self.web3.eth.send_raw_transaction(signed_txn.rawTransaction)
            # we wait for the transaction to be mined
            self.wait_for_transaction(result)

        protocol_contract = self.get_contract("PROTOCOL")
        txn = protocol_contract.functions.deposit(
            self.public_key,
            subaccount_id,
            required_wei,
            asset_contract.address,
        ).build_transaction(
            {
                "from": self.public_key,
                "nonce": self.web3.eth.get_transaction_count(self.public_key),
            }
        )
        signed_txn = self.wallet.sign_transaction(txn)
        result = self.web3.eth.send_raw_transaction(signed_txn.rawTransaction)
        return self.wait_for_transaction(result)

    def wait_for_transaction(self, txn_hash, timeout=60):
        while True:
            if timeout == 0:
                raise Exception("Timeout")
            try:
                receipt = self.web3.eth.get_transaction_receipt(txn_hash)
                if receipt is not None:
                    break
            except TransactionNotFound:
                time.sleep(1)
            timeout -= 1
        return receipt["status"] == 1

    def get_contract_address(self, name: str):
        """
        Get the contract address for a specific asset.
        """
        return self.web3.to_checksum_address(CONTRACTS[self.env][name])

    def get_contract(self, name: str):
        """
        Get the contract for a specific asset.
        """
        abis = {
            "USDB": ERC_20_ABI,
            "PROTOCOL": PROTOCOL_ABI,
        }
        return self.web3.eth.contract(
            address=self.get_contract_address(name),
            abi=abis[name],
        )



"""
Constants for the hundred_x package.
"""

from hundred_x.enums import ApiType, Environment

DEVNET_REST_URL = "https://api.dev.ciaobella.dev"
DEVNET_WEBSOCKET_URL = "https://stream.dev.ciaobella.dev"

TESTNET_REST_URL = "https://api.staging.100x.finance"
TESTNET_WEBSOCKET_URL = "https://stream.staging.100x.finance"

MAINNET_REST_URL = "https://api.100x.finance"
MAINNET_WEBSOCKET_URL = "https://stream.100x.finance"


APIS = {
    Environment.TESTNET: {
        ApiType.REST: TESTNET_REST_URL,
        ApiType.WEBSOCKET: TESTNET_WEBSOCKET_URL,
    },
    Environment.PROD: {
        ApiType.REST: MAINNET_REST_URL,
        ApiType.WEBSOCKET: MAINNET_WEBSOCKET_URL,
    },
    Environment.DEVNET: {
        ApiType.REST: DEVNET_REST_URL,
        ApiType.WEBSOCKET: DEVNET_WEBSOCKET_URL,
    },
}


LOGIN_MESSAGE = "I would like to login to 100x finance."

CONTRACTS = {
    Environment.PROD: {
        "USDB": "0x4300000000000000000000000000000000000003",
        "PROTOCOL": "0x1BaEbEE6B00B3f559B0Ff0719B47E0aF22A6bfC4",
        "VERIFYING_CONTRACT": "0x691a5fc3a81a144e36c6C4fBCa1fC82843c80d0d",
        "CHAIN_ID": 81457,
    },
    Environment.TESTNET: {
        "USDB": "0x79A59c326C715AC2d31C169C85d1232319E341ce",
        "PROTOCOL": "0x0c3b9472b3923CfE199bAE24B5f5bD75FAD2bae9",
        "VERIFYING_CONTRACT": "0x02Ca4fcB63E2D3C89fa20D86ccDcfc540c683545",
        "CHAIN_ID": 168587773,
    },
    Environment.DEVNET: {
        "USDB": "0x79A59c326C715AC2d31C169C85d1232319E341ce",
        "PROTOCOL": "0x9E4d21FeFE19EbD9ce94b71A60bce6CD793DD2BF",
        "VERIFYING_CONTRACT": "0x8ffdD4755aa383cD4CB715ABD8e1375926123729",
        "CHAIN_ID": 168587773,
    },
}



"""
Enmum for the hundred_x package
"""

from enum import Enum


class ApiType(Enum):
    """
    Enum for the API type.
    """

    REST = "REST"
    WEBSOCKET = "WEBSOCKET"


class Environment(Enum):
    """
    Enum for the environment of the client.
    """

    PROD = "prod"
    TESTNET = "testnet"
    DEVNET = "local"


class OrderType(Enum):
    """
    Enum for the order type.
    """

    LIMIT = 0
    LIMIT_MAKER = 1
    MARKET = 2
    STOP_LOSS = 3
    STOP_LOSS_LIMIT = 4
    TAKE_PROFIT = 5
    TAKE_PROFIT_LIMIT = 6


class TimeInForce(Enum):
    """
    Enum for the time in force.
    """

    GTC = 0
    FOK = 1
    IOC = 2


class OrderSide(Enum):
    """
    Enum for the order side.
    """

    BUY = True
    SELL = False



"""
Utils module for hundred_x package.
"""

import json
import os
from pathlib import Path
from typing import Any, Dict

from hundred_x.enums import Environment

INSTALL_DIR = os.path.dirname(os.path.abspath(__file__))
DEFAULT_ENCODING = "utf-8"

STRING_KEYS = [
    "price",
    "quantity",
]


def get_base_url(envirnment: Environment):
    """Map the environment to the corresponding base URL."""


def get_abi(contract_name: str):
    """Get the ABI of the contract."""
    with open(Path(INSTALL_DIR) / "abis" / f"{contract_name}.json", "r", encoding=DEFAULT_ENCODING) as f:
        return json.load(f)


def from_message_to_payload(message: Dict[str, Any]):
    """Convert a message to a payload."""
    for key, value in message.items():
        if key in STRING_KEYS:
            message[key] = str(value)
    return message



"""
Eip structs for the hundred_x project

"""

from eip712_structs import Address, Boolean, EIP712Struct, String, Uint


class LoginMessage(EIP712Struct):
    account = Address()
    message = String()
    timestamp = Uint(64)


class Withdraw(EIP712Struct):
    account = Address()
    subAccountId = Uint(8)
    asset = Address()
    quantity = Uint(128)
    nonce = Uint(64)


class Order(EIP712Struct):
    account = Address()
    subAccountId = Uint(8)
    productId = Uint(32)
    isBuy = Boolean()
    orderType = Uint(8)
    timeInForce = Uint(8)
    expiration = Uint(64)
    price = Uint(128)
    quantity = Uint(128)
    nonce = Uint(64)


class CancelOrder(EIP712Struct):
    account = Address()
    subAccountId = Uint(8)
    productId = Uint(32)
    orderId = String()


class CancelOrders(EIP712Struct):
    account = Address()
    subAccountId = Uint(8)
    productId = Uint(32)



"""
Async client for the HundredX API
"""
from hundred_x.client import HundredXClient


class AsyncHundredXClient(HundredXClient):
    """
    Asynchronous client for the HundredX API.
    """

    async def list_products(self):
        """
        List all products available on the exchange.
        """
        return super().list_products()

    async def get_product(self, symbol: str):
        """
        Get a specific product available on the exchange.
        """
        return super().get_product(symbol)

    async def get_server_time(self):
        """
        Get the server time.
        """
        return super().get_server_time()
    

